/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "common/setup_before.h"
#include "common/w3xp_protocol.h"
#include "bnpmap.h"
#include "common/eventlog.h"
#include "common/setup_after.h"

static const pmap_row pmap_table_w3xp_b301[] = {
     { CLIENT_COUNTRYINFO_109, CLIENT_W3XP_COUNTRYINFO },
     { SERVER_ECHOREQ, SERVER_W3XP_ECHOREQ },
     { CLIENT_ECHOREPLY, CLIENT_W3XP_ECHOREPLY },
     { SERVER_AUTHREQ_109, SERVER_W3XP_AUTHREQ },
     { CLIENT_AUTHREQ_109, CLIENT_W3XP_AUTHREQ },
     { SERVER_AUTHREPLY_109, SERVER_W3XP_AUTHREPLY},
     { CLIENT_FILEINFOREQ, CLIENT_W3XP_FILEINFOREQ },
     { SERVER_FILEINFOREPLY, SERVER_W3XP_FILEINFOREPLY },
     { CLIENT_ICONREQ, CLIENT_W3XP_ICONREQ },
     { SERVER_ICONREPLY, SERVER_W3XP_ICONREPLY },
     { CLIENT_LOGINREQ_W3, CLIENT_W3XP_LOGINREQ }, 
     { SERVER_LOGINREPLY_W3, SERVER_W3XP_LOGINREPLY },
     { CLIENT_LOGONPROOFREQ, SERVER_W3XP_LOGONPROOFREPLY },
     { SERVER_LOGONPROOFREPLY, SERVER_W3XP_LOGONPROOFREPLY },
     { CLIENT_CHANGEGAMEPORT, CLIENT_W3XP_CHANGEGAMEPORT },
     { CLIENT_PLAYERINFOREQ, CLIENT_W3XP_PLAYERINFOREQ },
     { SERVER_PLAYERINFOREPLY, CLIENT_W3XP_PLAYERINFOREQ },
     { CLIENT_MOTD_W3, CLIENT_W3XP_MOTD },
     { SERVER_MOTD_W3, SERVER_W3XP_MOTD },
     { CLIENT_JOINCHANNEL, CLIENT_W3XP_JOINCHANNEL },
     { SERVER_MESSAGE, SERVER_W3XP_MESSAGE },
     { CLIENT_STATSREQ, CLIENT_W3XP_STATSREQ },
     { SERVER_STATSREPLY, SERVER_W3XP_STATSREPLY },
     { CLIENT_FINDANONGAME, CLIENT_W3XP_ANONGAME_REQ },
     { SERVER_FINDANONGAME, SERVER_W3XP_ANONGAME_REPLY },
     { CLIENT_CREATEACCOUNT_W3, CLIENT_W3XP_CREATEACCOUNT },
     { SERVER_CREATEACCOUNT_W3, SERVER_W3XP_CREATEACCOUNT },
     { CLIENT_GAMELISTREQ, CLIENT_W3XP_GAMELISTREQ },
     { SERVER_GAMELISTREPLY, SERVER_W3XP_GAMELISTREPLY },
     { CLIENT_STARTGAME4, CLIENT_W3XP_STARTGAME },
     { CLIENT_JOIN_GAME, CLIENT_W3XP_JOIN_GAME },
     { CLIENT_MESSAGE, CLIENT_W3XP_MESSAGE },
     { SERVER_FRIENDADD_ACK, SERVER_W3XP_FRIENDADD_ACK },
     { CLIENT_FRIENDSLISTREQ, CLIENT_W3XP_FRIENDSLIST_REQ },
     { SERVER_FRIENDSLISTREPLY, SERVER_W3XP_FRIENDSLIST_REPLY },
     { SERVER_FRIENDDEL_ACK, SERVER_W3XP_FRIENDDEL_ACK },
     { CLIENT_FRIENDINFOREQ, CLIENT_W3XP_FRIENDINFO_REQ },
     { SERVER_FRIENDINFOREPLY, SERVER_W3XP_FRIENDINFO_REPLY },
     { CLIENT_ADREQ, CLIENT_W3XP_ADREQ },
     { SERVER_ADREPLY, SERVER_W3XP_ADREPLY },
     { CLIENT_ADCLICK2, 0x2dff },
     { CLIENT_W3ROUTE_REQ, CLIENT_W3XP_W3ROUTE_REQ },
     { SERVER_W3ROUTE_PLAYERINFO, SERVER_W3XP_W3ROUTE_PLAYERINFO },
     { CLIENT_W3ROUTE_GAMERESULT, CLIENT_W3XP_W3ROUTE_GAMERESULT },
     { CLIENT_PROGIDENT2, CLIENT_W3XP_CHANLISTREQ },
     { SERVER_CHANNELLIST, SERVER_W3XP_CHANLISTREPLY },
     { CLIENT_ARRANGEDTEAM_FRIENDSCREEN, CLIENT_W3XP_ARRANGEDTEAM_FRIENDSCREEN },
     { SERVER_ARRANGEDTEAM_FRIENDSCREEN, SERVER_W3XP_ARRANGEDTEAM_FRIENDSCREEN },
     { CLIENT_W3XP_CLAN_INFOREQ, CLIENT_W3XP_CLAN_INFOREQ },
     { SERVER_W3XP_CLAN_INFOREPLY, SERVER_W3XP_CLAN_INFOREPLY },
     { 0, 0}
};

static const pmap_row pmap_table_w3xp_b303[] = {
     { CLIENT_COUNTRYINFO_109, CLIENT_W3XP_COUNTRYINFO },
     { SERVER_ECHOREQ, SERVER_W3XP_ECHOREQ },
     { CLIENT_ECHOREPLY, CLIENT_W3XP_ECHOREPLY },
     { SERVER_AUTHREQ_109, SERVER_W3XP_AUTHREQ },
     { CLIENT_AUTHREQ_109, CLIENT_W3XP_AUTHREQ },
     { SERVER_AUTHREPLY_109, SERVER_W3XP_AUTHREPLY},
     { CLIENT_FILEINFOREQ, 0x01ff },
     { SERVER_FILEINFOREPLY, 0x01ff },
     { CLIENT_ICONREQ, 0x36ff },
     { SERVER_ICONREPLY, 0x36ff },
     { SERVER_CREATEACCOUNT_W3, 0x1cff },
     { CLIENT_LOGINREQ_W3, 0x79ff }, 
     { SERVER_LOGINREPLY_W3, 0x79ff },
     { CLIENT_LOGONPROOFREQ, 0x64ff },
     { SERVER_LOGONPROOFREPLY, 0x64ff },
     { CLIENT_CHANGEGAMEPORT, 0x60ff },
     { CLIENT_PLAYERINFOREQ, 0x65ff },
     { SERVER_PLAYERINFOREPLY, 0x65ff },
     { CLIENT_MOTD_W3, 0x42ff },
     { SERVER_MOTD_W3, 0x42ff },
     { CLIENT_PROGIDENT2, 0x71ff },
     { SERVER_CHANNELLIST, 0x71ff },
     { CLIENT_JOINCHANNEL, 0x34ff },
     { SERVER_MESSAGE, 0x0bff },
     { CLIENT_MESSAGE, 0x2eff },
     { CLIENT_STATSREQ, 0x03ff },
     { SERVER_STATSREPLY, 0x03ff },
     { CLIENT_FINDANONGAME, 0x16ff },
     { SERVER_FINDANONGAME, 0x16ff },
     { CLIENT_ADREQ, 0x7cff },
     { SERVER_ADREPLY, 0x7cff },
     { CLIENT_ADCLICK2, 0x57ff },
     { CLIENT_GAMELISTREQ, 0x44ff },
     { SERVER_GAMELISTREPLY, 0x44ff },
     { CLIENT_STARTGAME4, 0x24ff },
     { CLIENT_JOIN_GAME, 0x2aff },
     { CLIENT_CLOSEGAME, 0x58ff },
     { CLIENT_CLOSEGAME2, 0x0aff },
     { SERVER_FRIENDADD_ACK, 0x18ff },
     { CLIENT_FRIENDSLISTREQ, 0x56ff },
     { SERVER_FRIENDSLISTREPLY, 0x56ff },
     { SERVER_FRIENDDEL_ACK, 0x3cff },
     { CLIENT_W3XP_CLAN_INFOREQ, 0x55ff },
     { SERVER_W3XP_CLAN_INFOREPLY, 0x55ff },
     { CLIENT_FRIENDINFOREQ, 0x59ff },
     { SERVER_FRIENDINFOREPLY, 0x59ff },
     { CLIENT_ARRANGEDTEAM_FRIENDSCREEN, 0x5bff },
     { SERVER_ARRANGEDTEAM_FRIENDSCREEN, 0x5bff }, 
     { CLIENT_ARRANGEDTEAM_INVITE_FRIEND, 0x73ff },
     { SERVER_ARRANGEDTEAM_INVITE_FRIEND_ACK, 0x73ff },
     { SERVER_ARRANGEDTEAM_SEND_INVITE, 0x09ff },
     { CLIENT_ARRANGEDTEAM_ACCEPT_DECLINE_INVITE, 0x09ff },
     { SERVER_ARRANGEDTEAM_MEMBER_DECLINE, 0x69ff },
     { CLIENT_STATSUPDATE, 0x76ff },
     { CLIENT_NULL, 0x40ff },
     { 0, 0}
};


extern unsigned int bnpmap_real(unsigned int show, pmap_row const * table)
{
   pmap_row const * p;

   for(p = table; p->real != 0; p++)
     if (p->show == show) return p->real;

   eventlog(eventlog_level_error, __FUNCTION__, "unknown packet [0x%04X], please debug", show);
   return 0;
}

extern unsigned int bnpmap_show(unsigned int real, pmap_row const * table)
{
   pmap_row const * p;

   for(p = table; p->real != 0; p++)
     if (p->real == real) return p->show;

   eventlog(eventlog_level_error, __FUNCTION__, "unknown packet [0x%04X], please debug", real);
   return 0;
}

extern pmap_row const * bnpmap_get_war3xptable(unsigned long versionid)
{
   eventlog(eventlog_level_trace, __FUNCTION__,"versionid: %d", versionid);
   switch(versionid) {
   case 301:
      return pmap_table_w3xp_b301;
   case 303:
   case 304:
   case 305:
   case 306:
   case 307:
   case 308:
   case 309:
      return pmap_table_w3xp_b303;
   default:
      eventlog(eventlog_level_error, __FUNCTION__, "unknown war3xp versionid [%d]", versionid);
      return pmap_table_w3xp_b303;
   }
}
