/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#define CHARACTER_INTERNAL_ACCESS
#include "common/setup_before.h"
#include <stdio.h>
#ifdef HAVE_STDDEF_H
# include <stddef.h>
#else
# ifndef NULL
#  define NULL ((void *)0)
# endif
#endif
#ifdef STDC_HEADERS
# include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
# include <string.h>
#else
# ifdef HAVE_STRINGS_H
#  include <strings.h>
# endif
#endif
#include <errno.h>
#include <compat/strerror.h>
#include "common/eventlog.h"
#include "common/util.h"
#include "common/bnettime.h"
#include "ladder.h"
#include "account.h"
#include "character.h"
#include "connection.h"
#include "compat/strcasecmp.h"
#include "compat/strncasecmp.h"
#include "common/hashtable.h"
#include "common/setup_after.h"
#include "common/bnet_protocol.h"
#include "war3ladder.h"
#include "prefs.h"
#include "common/tag.h"

#define MaxRankKeptInLadder 1000

const char * WAR3_solo_file = "WAR3_solo";
const char * W3XP_solo_file = "W3XP_solo";
const char * WAR3_team_file = "WAR3_team";
const char * W3XP_team_file = "W3XP_team";
const char * WAR3_ffa_file  = "WAR3_ffa";
const char * W3XP_ffa_file  = "W3XP_ffa";
const char * WAR3_at_file   = "WAR3_atteam";
const char * W3XP_at_file   = "W3XP_atteam";
const char * std_end   = ".dat";
const char * xml_end   = ".xml";

char * WAR3_solo_filename, * WAR3_team_filename, * WAR3_ffa_filename, * WAR3_at_filename;
char * W3XP_solo_filename, * W3XP_team_filename, * W3XP_ffa_filename, * W3XP_at_filename;

int in_same_team(t_account * acc1, t_account * acc2, unsigned int tc1, unsigned int tc2, char const * clienttag)
{
   char const * teammembers1;
   char const * teammembers2;
   teammembers1 = account_get_atteammembers(acc1,tc1,clienttag);
   teammembers2 = account_get_atteammembers(acc2,tc2,clienttag);
   if ((teammembers1 == NULL) || (teammembers2 == NULL))
   {
      if (teammembers1 != NULL) return 1; else return 0;
   }
   if (strcmp(teammembers1, teammembers2)==0) return 1;
   else return 0;
}

extern t_w3_ladder * solo_ladder(char const * clienttag)
{
  if (strcmp(clienttag,CLIENTTAG_WARCRAFT3)==0) 
    return &WAR3_solo_ladder;
  else 
    return &W3XP_solo_ladder;
}

extern t_w3_ladder * team_ladder(char const * clienttag)
{
  if (strcmp(clienttag,CLIENTTAG_WARCRAFT3)==0) 
    return &WAR3_team_ladder;
  else 
    return &W3XP_team_ladder;
}

extern t_w3_ladder * ffa_ladder(char const * clienttag)
{
  if (strcmp(clienttag,CLIENTTAG_WARCRAFT3)==0) 
    return &WAR3_ffa_ladder;
  else 
    return &W3XP_ffa_ladder;
}

extern t_w3_ladder * at_ladder(char const * clienttag)
{
  if (strcmp(clienttag,CLIENTTAG_WARCRAFT3)==0) 
    return &WAR3_at_ladder;
  else 
    return &W3XP_at_ladder;
}

 extern int war3_ladder_add(t_w3_ladder *ladder, int uid, int xp, int rank, t_account *account, unsigned int teamcount,char const * clienttag)
  {

   t_w3_ladder_internal *ladder_entry;
   ladder_entry = malloc(sizeof(t_w3_ladder_internal));

   if (ladder_entry != NULL)
   {
      ladder_entry->uid       = uid;
      ladder_entry->xp        = xp;
      ladder_entry->rank      = rank;
      ladder_entry->account   = account;
      ladder_entry->teamcount = teamcount;
      ladder_entry->prev      = NULL;
      ladder_entry->next      = NULL;

      if (ladder->first == NULL)
      { // ladder empty, so just insert element
         ladder->first = ladder_entry;
         ladder->last  = ladder_entry;
      }
      else
      { // already elements in list
        // determine if first or last user in ladder has XP
        // closest to current, cause should be the shorter way to insert
        // new user
            if ((ladder->first->xp - xp) >= (xp - ladder->last->xp))
            // we should enter from last to first
            {
              t_w3_ladder_internal *search;
              search = ladder->last;
              while ((search != NULL) && (search->xp < xp))
              {
                search = search->next;
              }
	      // now we have to see if we have to correct position according to old ranking
	      if (rank!=0)
	      {
	          while ((search != NULL) && (search->xp == xp) && (search->rank > rank))
	          {
		    search = search->next;
	          }
	      }

	      if (teamcount!=0) // this only happens for atteams
		{
		  t_w3_ladder_internal *teamsearch;
		  t_w3_ladder_internal *teamfound;

		  teamsearch = search;
		  teamfound = NULL;

		  while ((teamsearch != NULL) && (teamsearch->xp == xp))
		    {
		      if (in_same_team(account,teamsearch->account,teamcount,teamsearch->teamcount,clienttag))
			{
			  teamfound = teamsearch;
			  break;
			}
		      teamsearch = teamsearch->next;
		    }
		  if (teamfound!=NULL) search = teamfound;

		}

              if (search == NULL)
              {
                ladder->first->next = ladder_entry;
                ladder_entry->prev  = ladder->first;
                ladder->first       = ladder_entry;
              }
              else
              {
                 ladder_entry->next = search;
                 ladder_entry->prev = search->prev;
                 if (search == ladder->last)
                 // enter at end of list
                 {
                  search->prev = ladder_entry;
                  ladder->last = ladder_entry;
                 }
                 else
                 {
                   search->prev->next = ladder_entry;
                   search->prev       = ladder_entry;
                 }
              }
            }
            else
            // start from first and the search towards last
            {
              t_w3_ladder_internal *search;
              search = ladder->first;
              while ((search != NULL) && (search->xp > xp))
              {
                search = search->prev;
              }
	      if (rank!=0)
	      {
	          while ((search != NULL) && (search->xp == xp) && (search->rank < rank))
	          {
                    search = search->prev;
	          }
	      }

	      if (teamcount!=0) // this only happens for atteams
		{
		  t_w3_ladder_internal *teamsearch;
		  t_w3_ladder_internal *teamfound;

		  teamsearch = search;
		  teamfound = NULL;

		  while ((teamsearch != NULL) && (teamsearch->xp == xp))
		    {
		      if (in_same_team(account,teamsearch->account,teamcount,teamsearch->teamcount,clienttag))
			{
			  teamfound = teamsearch;
			  break;
			}
		      teamsearch = teamsearch->prev;
		    }
		  if (teamfound!=NULL) search = teamfound;

		}


              if (search == NULL)
              {
                ladder->last->prev = ladder_entry;
                ladder_entry->next = ladder->last;
                ladder->last       = ladder_entry;
              }
              else
              {
                 ladder_entry->prev = search;
                 ladder_entry->next = search->next;
                 if (search == ladder->first)
                 // enter at beginning of list
                 {
                  search->next  = ladder_entry;
                  ladder->first = ladder_entry;
                 }
                 else
                 {
                   search->next->prev = ladder_entry;
                   search->next       = ladder_entry;
                 }
              }
            }
       }
       ladder->dirty = 1;
       return 0;
   }
   else
   {
     eventlog(eventlog_level_error,"war3_ladder_add","could not add account to ladder");
     return -1;
   }
 }

 extern int war3_ladder_update(t_w3_ladder *ladder, int uid, int xp, t_account *account, unsigned int teamcount, char const * clienttag)
 {
   t_w3_ladder_internal *search, *newpos;
   search = ladder->first;
   while (search && ((search->uid != uid) || (search->teamcount!=teamcount))) { search = search->prev; }
   if (search != NULL)
   {
     search->xp += xp;

     // make sure, we don't get negative XP
     if (search->xp < 0) search->xp = 0;

     if (xp>0)
     // XP gained.... maybe getting better ranking
     {
        newpos = search->next;
        while ((newpos != NULL) && (newpos->xp < search->xp))
        {
          newpos = newpos->next;
        }
        if (newpos != search->next)
        // so we really have to change ranking now
        {
	  // we are changing something in the ladder
          ladder->dirty = 1;
          // first close gap, where we've been...
          search->next->prev = search->prev;
          if (search->prev != NULL) search->prev->next = search->next;
          // and then move to new position
          search->next = newpos;
          if (newpos == NULL)
          {
             search->prev = ladder->first;
             ladder->first->next = search;
             ladder->first       = search;
          }
          else
          {
            search->prev = newpos->prev;
            newpos->prev->next = search;
            newpos->prev = search;
          }
        }
     }
     if (xp<0)
     // XP lost.... maybe ranking gets worse
     {
        newpos = search->prev;
        while ((newpos != NULL) && (newpos->xp > search->xp))
        {
          newpos = newpos->prev;
        }
        if (newpos != search->prev)
        // so we really have to change ranking now
        {
	  // we are changing something in the ladder
          ladder->dirty = 1;
          // first close gap, where we've been...
          search->prev->next = search->next;
          if (search->next != NULL) search->next->prev = search->prev;
          // and then move to new position
          search->prev = newpos;
          if (newpos == NULL)
          {
             search->next = ladder->last;
             ladder->last->prev = search;
             ladder->last       = search;
          }
          else
          {
            search->next = newpos->next;
            newpos->next->prev = search;
            newpos->next = search;
          }
        }
     }
     return 0;
   }
   else
   { // account not yet in ladder... so... add it now...
     if (xp<0) xp=0;
     return war3_ladder_add(ladder,uid,xp,0,account,teamcount,clienttag);
   }
 }

 extern int war3_ladder_get_rank(t_w3_ladder *ladder, int uid, unsigned int teamcount, char const * clienttag)
 {
   int ranking = 1;
   t_w3_ladder_internal *search;
   search = ladder->first;
   while ((search!=NULL) && ((search->uid != uid) || (search->teamcount!=teamcount))) 
   { 
     search = search->prev; 
     ranking++;
     if (ladder==at_ladder(clienttag))
     {
       // if in the same team as previous account
       if ((search) && (search->next) 
           && in_same_team(search->account,search->next->account,teamcount,search->next->teamcount,clienttag))
	       ranking--;
     }
   }
   if (search != NULL)
   {
     return ranking;
   }
   else
   {
     return 0;
   }
 }

extern t_account * war3_ladder_get_account(t_w3_ladder *ladder, int rank, unsigned int * teamcount,char const * clienttag)
{
  int ranking = 1;
  t_w3_ladder_internal *search;
  search = ladder->first;
  while ((search!=NULL) && (ranking<rank)) 
  { 
     search = search->prev; 
     ranking++;
     if (ladder == at_ladder(clienttag))
     {
       if ((search) && (search->next)
	  && in_same_team(search->account,search->next->account,search->teamcount,search->next->teamcount,clienttag))
	      ranking--;
     }
  }
  if (search) 
  {
    *teamcount = search->teamcount;
    return search->account;
  }
  else
  {
    *teamcount = 0;
    return NULL;
  }
}

extern int war3_ladder_update_accounts(t_w3_ladder *ladder, int (*set_fct)(), int (*get_fct1)(), int (*get_fct2)(),char const * clienttag)
{
    t_w3_ladder_internal *pointer, *tmp_pointer;
    t_account *account;
    int rank = 1;
    int update = 0;

    if (ladder->dirty == 1)
    {
      pointer = ladder->first;
      while (pointer!=NULL)
      {
         account = pointer->account;
         if (rank <= MaxRankKeptInLadder) 
         {
           // if account_get_rank and account_get_rank_ladder return different results, set rank
           if ((ladder!=at_ladder(clienttag)))
	   {
             if ((*get_fct1)(account,clienttag)!=(*get_fct2)(account,clienttag)) 
 	       {	
	         (*set_fct)(account,clienttag,rank);
	         update++;	
	       }
	   }
	   else
	   {
	     if ((*get_fct1)(account,pointer->teamcount,clienttag)!=(*get_fct2)(account,pointer->teamcount,clienttag))
	     {
	       (*set_fct)(account,rank,pointer->teamcount,clienttag);
	       update++;
	       // if in the same team as previous account
	       if ((pointer) && (pointer->next) 
                   && in_same_team(pointer->account,pointer->next->account,pointer->teamcount,pointer->next->teamcount,clienttag))
	       rank--;
             }
	   }
	   pointer=pointer->prev;
	   rank++;
	  
         }
         else
	 {
	   // leave while loop
           break;
	 }
      }
      while (pointer!=NULL)
      {
        // all accounts following now are out of the ladder range we keep track of....
	// so set rank to 0 and remove account from ladder
	if (ladder!=at_ladder(clienttag))
	 { if ((*get_fct1)(account,clienttag)!=0) (*set_fct)(account,clienttag,0);}
	else
	 { if ((*get_fct1)(account,pointer->teamcount,clienttag)!=0) (*set_fct)(account,0,pointer->teamcount,clienttag);}
	// remove account from ladder
	if (pointer->next!=NULL) pointer->next->prev = pointer->prev;
	if (pointer->prev!=NULL) pointer->prev->next = pointer->next;
	if (ladder->last == pointer)  ladder->last  = pointer->next;
	if (ladder->first == pointer) ladder->first = pointer->prev;
	tmp_pointer = pointer->prev;
	free((void *)pointer);
	pointer = tmp_pointer;
      }
      eventlog(eventlog_level_info,"war3_ladder_update_accounts","updated %u accounts",update);
    }
    ladder->dirty = 0;
    return 0;
}

extern int war3_ladder_update_all_accounts(void)
{
  eventlog(eventlog_level_info,"war3_ladder_update_all_accounts","updating ranking for all accounts");
  war3_ladder_update_accounts(solo_ladder(CLIENTTAG_WARCRAFT3),&account_set_solorank_ladder,
                              &account_get_solorank, &account_get_solorank_ladder,CLIENTTAG_WARCRAFT3);
  war3_ladder_update_accounts(team_ladder(CLIENTTAG_WARCRAFT3),&account_set_teamrank_ladder,
                              &account_get_teamrank, &account_get_teamrank_ladder,CLIENTTAG_WARCRAFT3);
  war3_ladder_update_accounts(ffa_ladder(CLIENTTAG_WARCRAFT3),&account_set_ffarank_ladder,
		  	      &account_get_ffarank, &account_get_ffarank_ladder,CLIENTTAG_WARCRAFT3);
  war3_ladder_update_accounts(at_ladder(CLIENTTAG_WARCRAFT3), &account_set_atteamrank_ladder,
		  	      &account_get_atteamrank, &account_get_atteamrank_ladder,CLIENTTAG_WARCRAFT3);
  war3_ladder_update_accounts(solo_ladder(CLIENTTAG_WAR3XP),&account_set_solorank_ladder,
                              &account_get_solorank, &account_get_solorank_ladder,CLIENTTAG_WAR3XP);
  war3_ladder_update_accounts(team_ladder(CLIENTTAG_WAR3XP),&account_set_teamrank_ladder,
                              &account_get_teamrank, &account_get_teamrank_ladder,CLIENTTAG_WAR3XP);
  war3_ladder_update_accounts(ffa_ladder(CLIENTTAG_WAR3XP),&account_set_ffarank_ladder,
		  	      &account_get_ffarank, &account_get_ffarank_ladder,CLIENTTAG_WAR3XP);
  war3_ladder_update_accounts(at_ladder(CLIENTTAG_WAR3XP), &account_set_atteamrank_ladder,
		  	      &account_get_atteamrank, &account_get_atteamrank_ladder,CLIENTTAG_WAR3XP);
  return 0;
}

extern void war3_ladder_remove_account(t_w3_ladder * ladder, t_account * account)
{
  t_w3_ladder_internal * pointer, * tmp_pointer;
  pointer = ladder->first;
  while (pointer)
  {
    if (pointer->account == account) 
    {
	    // TODO remove acc from ladder
      if (pointer->prev)
      {
	pointer->prev->next = pointer->next;
      }
      else
      {
        ladder->last = pointer->next;
      }
      if (pointer->next)
      {
	pointer->next->prev = pointer->prev;
      }
      else
      {
         ladder->first = pointer->prev;
      }
      tmp_pointer = pointer;
      pointer = pointer->prev;
      free(tmp_pointer);
      ladder->dirty = 1;	    
    }
    else
      pointer=pointer->prev;
  }
}

extern void war3_ladders_remove_account(t_account * account)
{
  if (account_get_solorank(account,CLIENTTAG_WARCRAFT3))    
	  war3_ladder_remove_account(solo_ladder(CLIENTTAG_WARCRAFT3),account);
  if (account_get_teamrank(account,CLIENTTAG_WARCRAFT3))    
	  war3_ladder_remove_account(team_ladder(CLIENTTAG_WARCRAFT3),account);
  if (account_get_ffarank(account,CLIENTTAG_WARCRAFT3))     
	  war3_ladder_remove_account(ffa_ladder(CLIENTTAG_WARCRAFT3),account);
  if (account_get_atteamcount(account,CLIENTTAG_WARCRAFT3)) 
	  war3_ladder_remove_account(at_ladder(CLIENTTAG_WARCRAFT3),account);
  if (account_get_solorank(account,CLIENTTAG_WAR3XP))       
	  war3_ladder_remove_account(solo_ladder(CLIENTTAG_WAR3XP),account);
  if (account_get_teamrank(account,CLIENTTAG_WAR3XP))       
	  war3_ladder_remove_account(team_ladder(CLIENTTAG_WAR3XP),account);
  if (account_get_ffarank(account,CLIENTTAG_WAR3XP))        
	  war3_ladder_remove_account(ffa_ladder(CLIENTTAG_WAR3XP),account);
  if (account_get_atteamcount(account,CLIENTTAG_WAR3XP))    
	  war3_ladder_remove_account(at_ladder(CLIENTTAG_WAR3XP),account);
}

extern void war3_ladders_load_accounts_to_ladderlists(void)
{
  t_entry   * curr;
  t_account * account;
  int rank, teamcount;

  HASHTABLE_TRAVERSE(accountlist(),curr)
    {
      if (curr && (account=((t_account *)entry_get_data(curr))))
	{
	  if (account)
	    { 
	      int rank;
	      int uid = account_get_uid(account);
	      if ((rank = account_get_solorank(account,CLIENTTAG_WARCRAFT3)))
		{
		  war3_ladder_add(solo_ladder(CLIENTTAG_WARCRAFT3),
				  uid,
				  account_get_soloxp(account,CLIENTTAG_WARCRAFT3),
				  rank,
				  account,0,CLIENTTAG_WARCRAFT3);
		}
	      if ((rank = account_get_teamrank(account,CLIENTTAG_WARCRAFT3)))
		{
		  war3_ladder_add(team_ladder(CLIENTTAG_WARCRAFT3),
				  uid,
				  account_get_teamxp(account,CLIENTTAG_WARCRAFT3),
				  rank,
				  account,0,CLIENTTAG_WARCRAFT3 );

		}
	      if ((rank = account_get_ffarank(account,CLIENTTAG_WARCRAFT3)))
		{
		  war3_ladder_add(ffa_ladder(CLIENTTAG_WARCRAFT3),
				  uid,
				  account_get_ffaxp(account,CLIENTTAG_WARCRAFT3),
				  rank,
				  account,0,CLIENTTAG_WARCRAFT3 );
		}
	      // user is part of a team
	      if (teamcount = account_get_atteamcount(account,CLIENTTAG_WARCRAFT3))
		{
		  int counter, rank;
		  // for each team he is in...
		  for (counter=1; counter<=teamcount; counter++)
		    {
		      rank = account_get_atteamrank(account,counter,CLIENTTAG_WARCRAFT3);
		      // make sure... user has ranking... and team is valid...
		      if ((rank) && account_get_atteammembers(account,counter,CLIENTTAG_WARCRAFT3))
			{
			  war3_ladder_add(at_ladder(CLIENTTAG_WARCRAFT3),
					  uid,
					  account_get_atteamxp(account,counter,CLIENTTAG_WARCRAFT3),
					  rank,
					  account,
					  counter,CLIENTTAG_WARCRAFT3);
			}
		    }	   
		}
	      if ((rank = account_get_solorank(account,CLIENTTAG_WAR3XP)))
		{
		  war3_ladder_add(solo_ladder(CLIENTTAG_WAR3XP),
				  uid,
				  account_get_soloxp(account,CLIENTTAG_WAR3XP),
				  rank,
				  account,0,CLIENTTAG_WAR3XP);
		}
	      if ((rank = account_get_teamrank(account,CLIENTTAG_WAR3XP)))
		{
		  war3_ladder_add(team_ladder(CLIENTTAG_WAR3XP),
				  uid,
				  account_get_teamxp(account,CLIENTTAG_WAR3XP),
				  rank,
				  account,0,CLIENTTAG_WAR3XP );

		}
	      if ((rank = account_get_ffarank(account,CLIENTTAG_WAR3XP)))
		{
		  war3_ladder_add(ffa_ladder(CLIENTTAG_WAR3XP),
				  uid,
				  account_get_ffaxp(account,CLIENTTAG_WAR3XP),
				  rank,
				  account,0,CLIENTTAG_WAR3XP );
		}
	      // user is part of a team
	      if (teamcount = account_get_atteamcount(account,CLIENTTAG_WAR3XP))
		{
		  int counter, rank;
		  // for each team he is in...
		  for (counter=1; counter<=teamcount; counter++)
		    {
		      rank = account_get_atteamrank(account,counter,CLIENTTAG_WAR3XP);
		      // make sure... user has ranking... and team is valid...
		      if ((rank) && account_get_atteammembers(account,counter,CLIENTTAG_WAR3XP))
			{
			  war3_ladder_add(at_ladder(CLIENTTAG_WAR3XP),
					  uid,
					  account_get_atteamxp(account,counter,CLIENTTAG_WAR3XP),
					  rank,
					  account,
					  counter,CLIENTTAG_WAR3XP);
			}
		    }	   
		}
	    }
	}
    }
}


int standard_writer(FILE * fp, t_w3_ladder * ladder,char const * clienttag)
{
  t_w3_ladder_internal * pointer;
  unsigned int rank=0;

  pointer = ladder->first;
  while (pointer != NULL)
  {
     rank++;

     if (ladder==at_ladder(clienttag))
     {
       // if in the same team as previous account
       if ((pointer) && (pointer->next) 
           && in_same_team(pointer->account,pointer->next->account,pointer->teamcount,pointer->next->teamcount,clienttag))
	       rank--;
       else
	 // other team... so write all team members names, xp and rank to file
	 fprintf(fp,"%s,%u,%u\n",account_get_atteammembers(pointer->account,pointer->teamcount,clienttag),pointer->xp,rank);
     }
     else
     // write username, xp and rank to file
     fprintf(fp,"%s,%u,%u\n",account_get_name(pointer->account),pointer->xp,rank);

     pointer=pointer->prev;
  }
  return 0;
}

int XML_writer(FILE * fp, t_w3_ladder * ladder, char const * clienttag)
  /* XML Ladder files
   * added by jfro 
   * 1/2/2003
   */
{
  t_w3_ladder_internal * pointer;
  unsigned int rank=0;
  unsigned int level=0;
  unsigned int wins=0;
  unsigned int losses=0;
  unsigned int orc_wins,human_wins,undead_wins,nightelf_wins,random_wins;
  unsigned int orc_losses,human_losses,undead_losses,nightelf_losses,random_losses;
  char *members=NULL;
  char *member;

  fprintf(fp,"<?xml version=\"1.0\"?>\n<ladder>\n");
  pointer = ladder->first;
  while (pointer != NULL)
  {
     rank++;

     if (ladder==at_ladder(clienttag))
     {
       // if in the same team as previous account
       if ((pointer) && (pointer->next) 
           && in_same_team(pointer->account,pointer->next->account,pointer->teamcount,pointer->next->teamcount,clienttag))
	       rank--;
       else
	 {
	   // other team... so write all team members names, xp and rank to file
	   fprintf(fp,"\t<team>\n");
	   if (account_get_atteammembers(pointer->account,pointer->teamcount,clienttag)==NULL)
	     {
	       eventlog(eventlog_level_error,"XML_writer","got invalid team, skipping");
	       pointer=pointer->prev;
	       continue;
	     }
	   
	   if (members = strdup(account_get_atteammembers(pointer->account,pointer->teamcount,clienttag)))
	     {
	       for ( member = strtok(members," ");
		     member;
		     member = strtok(NULL," "))
		 fprintf(fp,"\t\t<member>%s</member>\n",member);
	       free(members);
	       fprintf(fp,"\t\t<xp>%u</xp>\n\t\t<rank>%u</rank>\n\t</team>\n",pointer->xp,rank);
	     }
	 }
     }
     else {
         if (ladder==solo_ladder(clienttag)) {
           level = account_get_sololevel(pointer->account,clienttag);
           wins = account_get_solowin(pointer->account,clienttag);
           losses = account_get_sololoss(pointer->account,clienttag);
         }
         else if (ladder==team_ladder(clienttag)) {
           level = account_get_teamlevel(pointer->account,clienttag);
           wins = account_get_teamwin(pointer->account,clienttag);
           losses = account_get_teamloss(pointer->account,clienttag);
         }
         else if (ladder==ffa_ladder(clienttag)) {
           level = account_get_ffalevel(pointer->account,clienttag);
           wins = account_get_ffawin(pointer->account,clienttag);
           losses = account_get_ffaloss(pointer->account,clienttag);
         }
         else {
           level = 0;
           wins = 0;
           losses = 0;
         }
         orc_wins = account_get_racewin(pointer->account,W3_RACE_ORCS,clienttag);
         orc_losses = account_get_raceloss(pointer->account,W3_RACE_ORCS,clienttag);
         undead_wins = account_get_racewin(pointer->account,W3_RACE_UNDEAD,clienttag);
         undead_losses = account_get_raceloss(pointer->account,W3_RACE_UNDEAD,clienttag);
         human_wins = account_get_racewin(pointer->account,W3_RACE_HUMANS,clienttag);
         human_losses = account_get_raceloss(pointer->account,W3_RACE_HUMANS,clienttag);
         nightelf_wins = account_get_racewin(pointer->account,W3_RACE_NIGHTELVES,clienttag);
         nightelf_losses = account_get_raceloss(pointer->account,W3_RACE_NIGHTELVES,clienttag);
         random_wins = account_get_racewin(pointer->account,W3_RACE_RANDOM,clienttag);
         random_losses = account_get_raceloss(pointer->account,W3_RACE_RANDOM,clienttag);
     // write username, xp and rank to file and everyhing else needed for nice ladder pages
	 fprintf(fp,"\t<player>\n\t\t<name>%s</name>\n\t\t<level>%u</level>\n\t\t<xp>%u</xp>\n",
                    account_get_name(pointer->account),level,pointer->xp);
	 fprintf(fp,"\t\t<wins>%u</wins>\n\t\t<losses>%u</losses>\n\t\t<rank>%u</rank>\n",
                    wins,losses,rank);
	 fprintf(fp,"\t\t<races>\n\t\t\t<orc>\n\t\t\t\t<wins>%u</wins>\n\t\t\t\t<losses>%u</losses>\n\t\t\t</orc>\n",
                    orc_wins,orc_losses);
	 fprintf(fp,"\t\t\t<human>\n\t\t\t\t<wins>%u</wins>\n\t\t\t\t<losses>%u</losses>\n\t\t\t</human>\n",
                    human_wins,human_losses);
	 fprintf(fp,"\t\t\t<nightelf>\n\t\t\t\t<wins>%u</wins>\n\t\t\t\t<losses>%u</losses>\n\t\t\t</nightelf>\n",
		    nightelf_wins,nightelf_losses);
	 fprintf(fp,"\t\t\t<undead>\n\t\t\t\t<wins>%u</wins>\n\t\t\t\t<losses>%u</losses>\n\t\t\t</undead>\n",
		    undead_wins,undead_losses);
	 fprintf(fp,"\t\t\t<random>\n\t\t\t\t<wins>%u</wins>\n\t\t\t\t<losses>%u</losses>\n\t\t\t</random>\n\t\t</races>\n\t</player>\n",
                    random_wins,random_losses);
     }
     pointer=pointer->prev;
  }
  fprintf(fp,"</ladder>\n");
  return 0;
}


extern int war3_ladder_write_to_file(char const * filename, t_w3_ladder * ladder, char const * clienttag)
{
  FILE * fp;
  
  if (!filename)
  {
    eventlog(eventlog_level_error,"war3_ladder_write_to_file","got NULL filename");
    return -1;
  }
  
  if (!(fp = fopen(filename,"w")))
  { 
     eventlog(eventlog_level_error,"war3_ladder_write_to_file","could not open file \"%s\" for writing (fopen: %s)",filename,strerror(errno)); 
     return -1;
  }

  if (prefs_get_XML_output_ladder())
    XML_writer(fp,ladder,clienttag);
  else
    standard_writer(fp,ladder,clienttag);

  fclose(fp);
  return 0;
}

extern int war3_ladders_write_to_file()
{
  
  eventlog(eventlog_level_info,"war3_ladders_write_to_file","flushing ladders to disc");
  war3_ladder_write_to_file(WAR3_solo_filename, solo_ladder(CLIENTTAG_WARCRAFT3),CLIENTTAG_WARCRAFT3);
  war3_ladder_write_to_file(WAR3_team_filename, team_ladder(CLIENTTAG_WARCRAFT3),CLIENTTAG_WARCRAFT3);
  war3_ladder_write_to_file(WAR3_ffa_filename,  ffa_ladder(CLIENTTAG_WARCRAFT3), CLIENTTAG_WARCRAFT3);
  war3_ladder_write_to_file(WAR3_at_filename,   at_ladder(CLIENTTAG_WARCRAFT3),  CLIENTTAG_WARCRAFT3);
  war3_ladder_write_to_file(W3XP_solo_filename, solo_ladder(CLIENTTAG_WAR3XP),   CLIENTTAG_WAR3XP);
  war3_ladder_write_to_file(W3XP_team_filename, team_ladder(CLIENTTAG_WAR3XP),   CLIENTTAG_WAR3XP);
  war3_ladder_write_to_file(W3XP_ffa_filename,  ffa_ladder(CLIENTTAG_WAR3XP),    CLIENTTAG_WAR3XP);
  war3_ladder_write_to_file(W3XP_at_filename,   at_ladder(CLIENTTAG_WAR3XP),     CLIENTTAG_WAR3XP);
  return 0;
}

void war3_ladder_init(t_w3_ladder *ladder)
 {
    ladder->first = NULL;
    ladder->last  = NULL;
    ladder->dirty = 0;
 }

extern char * create_filename(const char * path, const char * filename, const char * ending)
{
  char * result;
  if (!(result = malloc(strlen(path)+1+strlen(filename)+strlen(ending)+1)))
  {
    eventlog(eventlog_level_error,"create_filename","could not allocate name for filename %s%s",filename,ending);
    return NULL;
  }
  else
  {
    sprintf(result,"%s/%s%s",path,filename,ending);
    return result;
  }
}

void dispose_filename(char * filename)
{
  if (filename) free(filename);
}

void create_filenames(void)
{
  if (prefs_get_XML_output_ladder())
  {
    WAR3_solo_filename = create_filename(prefs_get_ladderdir(),WAR3_solo_file,xml_end);
    WAR3_team_filename = create_filename(prefs_get_ladderdir(),WAR3_team_file,xml_end);
    WAR3_ffa_filename  = create_filename(prefs_get_ladderdir(),WAR3_ffa_file,xml_end);
    WAR3_at_filename   = create_filename(prefs_get_ladderdir(),WAR3_at_file,xml_end);
    W3XP_solo_filename = create_filename(prefs_get_ladderdir(),W3XP_solo_file,xml_end);
    W3XP_team_filename = create_filename(prefs_get_ladderdir(),W3XP_team_file,xml_end);
    W3XP_ffa_filename  = create_filename(prefs_get_ladderdir(),W3XP_ffa_file,xml_end);
    W3XP_at_filename   = create_filename(prefs_get_ladderdir(),W3XP_at_file,xml_end);
  }
  else
  {
    WAR3_solo_filename = create_filename(prefs_get_ladderdir(),WAR3_solo_file,std_end);
    WAR3_team_filename = create_filename(prefs_get_ladderdir(),WAR3_team_file,std_end);
    WAR3_ffa_filename  = create_filename(prefs_get_ladderdir(),WAR3_ffa_file,std_end);
    WAR3_at_filename   = create_filename(prefs_get_ladderdir(),WAR3_at_file,std_end);
    W3XP_solo_filename = create_filename(prefs_get_ladderdir(),W3XP_solo_file,std_end);
    W3XP_team_filename = create_filename(prefs_get_ladderdir(),W3XP_team_file,std_end);
    W3XP_ffa_filename  = create_filename(prefs_get_ladderdir(),W3XP_ffa_file,std_end);
    W3XP_at_filename   = create_filename(prefs_get_ladderdir(),W3XP_at_file,std_end);
  }
}

extern void war3_ladders_init(void)
{
  eventlog(eventlog_level_info,"war3_ladders_init","initializing war3 ladders");
  war3_ladder_init(solo_ladder(CLIENTTAG_WARCRAFT3));
  war3_ladder_init(team_ladder(CLIENTTAG_WARCRAFT3));
  war3_ladder_init(ffa_ladder(CLIENTTAG_WARCRAFT3));
  war3_ladder_init(at_ladder(CLIENTTAG_WARCRAFT3));
  war3_ladder_init(solo_ladder(CLIENTTAG_WAR3XP));
  war3_ladder_init(team_ladder(CLIENTTAG_WAR3XP));
  war3_ladder_init(ffa_ladder(CLIENTTAG_WAR3XP));
  war3_ladder_init(at_ladder(CLIENTTAG_WAR3XP));
  create_filenames();
}

void war3_ladder_destroy(t_w3_ladder *ladder)
{
  t_w3_ladder_internal *pointer;
  if (ladder->dirty)
  {
     eventlog(eventlog_level_error,"war3_ladder_destroy","cannot destroy dirty ladder!");
  }
  else
  {
    while (ladder->first!=NULL)
    {
      pointer = ladder->first;
      ladder->first = pointer->prev;
      free((void *)pointer);
    }
  }
}

void dispose_filenames(void)
{
  dispose_filename(WAR3_solo_filename);
  dispose_filename(WAR3_team_filename);
  dispose_filename(WAR3_ffa_filename);
  dispose_filename(WAR3_at_filename);
  dispose_filename(W3XP_solo_filename);
  dispose_filename(W3XP_team_filename);
  dispose_filename(W3XP_ffa_filename);
  dispose_filename(W3XP_at_filename);
}

extern void war3_ladders_destroy(void)
{
  eventlog(eventlog_level_info,"war3_ladders_destroy","destroying war3 ladders");
  war3_ladder_destroy(solo_ladder(CLIENTTAG_WARCRAFT3));
  war3_ladder_destroy(team_ladder(CLIENTTAG_WARCRAFT3));
  war3_ladder_destroy(ffa_ladder(CLIENTTAG_WARCRAFT3));
  war3_ladder_destroy(at_ladder(CLIENTTAG_WARCRAFT3));
  war3_ladder_destroy(solo_ladder(CLIENTTAG_WAR3XP));
  war3_ladder_destroy(team_ladder(CLIENTTAG_WAR3XP));
  war3_ladder_destroy(ffa_ladder(CLIENTTAG_WAR3XP));
  war3_ladder_destroy(at_ladder(CLIENTTAG_WAR3XP));
  dispose_filenames();
}

extern void war3_ladder_reload_conf(void)
{
  dispose_filenames();
  create_filenames();
}
