/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

 #ifndef NULL
 # define NULL ((void *)0)
 #endif
 #define JUST_NEED_TYPES
 #include "account.h"
 #undef JUST_NEED_TYPES

  typedef struct war3_ladder_internal
 {
   int uid;
   int xp;
   int rank;
   unsigned int teamcount;            // needed for AT ladder
   t_account *account;
   struct war3_ladder_internal *prev; // user with less XP
   struct war3_ladder_internal *next; // user with more XP
 } t_w3_ladder_internal;

 typedef struct war3_ladder
 {
    t_w3_ladder_internal *first;
    t_w3_ladder_internal *last;
    int dirty;                        // 0==no changes, 1==something changed
 } t_w3_ladder;

t_w3_ladder WAR3_solo_ladder, WAR3_team_ladder, WAR3_ffa_ladder, WAR3_at_ladder;
t_w3_ladder W3XP_solo_ladder, W3XP_team_ladder, W3XP_ffa_ladder, W3XP_at_ladder;

 extern int war3_ladder_add(t_w3_ladder *ladder, int uid, int xp, int rank, t_account *account, unsigned int teamcount,char const * clienttag);
 // this function adds a user to the ladder and keeps the ladder sorted
 // returns 0 if everything is fine and -1 when error occured

 extern int war3_ladder_update(t_w3_ladder *ladder, int uid, int xp, t_account *account, unsigned int teamcount, char const * clienttag);
 // this functions increases the xp of user with UID uid and corrects ranking
 // returns 0 if everything is fine
 // if user is not yet in ladder, he gets added automatically

 extern int war3_ladder_get_rank(t_w3_ladder *ladder, int uid, unsigned int teamcount, char const * clienttag);
 // this function returns the rank of a user with a given uid
 // returns 0 if no such user is found
 
 extern int war3_ladder_update_all_accounts(void);
 // write the correct ranking information to all user accounts
 // and cut down ladder size to given limit
 
 extern int war3_ladders_write_to_file(void);
 // outputs the ladders into  files - for the guys that wanna make ladder pages
 
 extern void war3_ladders_init(void);
 // initialize the ladders

 extern void war3_ladders_destroy(void);
 // remove all ladder data from memory
 
 extern void war3_ladders_load_accounts_to_ladderlists(void);
 // enters all accounts from accountlist into the ladders
 
 extern void war3_ladder_reload_conf(void);
 // reloads relevant parameters from bnetd.conf (xml/std mode for ladder)
 
 extern t_account * war3_ladder_get_account(t_w3_ladder *ladder,int rank, unsigned int * teamcount, char const * clienttag);
 // returns the account that is on specified rank in specified ladder. also return teamcount for AT ladder
 // returns NULL if this rank is still vacant
 
 extern void war3_ladders_remove_account(t_account * account);
 // removes an account from all war3 ladders
 
 extern t_w3_ladder * solo_ladder(char const * clienttag);
 extern t_w3_ladder * team_ladder(char const * clienttag);
 extern t_w3_ladder * ffa_ladder(char const * clienttag);
 extern t_w3_ladder * at_ladder(char const * clienttag);
 // for external and internal clienttag specific reference of the ladders

 extern char * create_filename(const char * path, const char * filename, const char * ending);
 // Add by DJP for output.c
